#include <sys/param.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/time.h>
#include <sys/signal.h>
#include <sys/poll.h>
#include <sys/types.h>


#include <netinet/in.h>
#include <netinet/ip.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <errno.h>

#ifdef GETOPTLONG
#include <getopt.h>
#endif

#define VERSION_MAJOR	0
#define VERSION_MINOR	1
#define VERSION_REV	3


extern int errno ; 

#define	READ_BUF_LEN	512

#define DEFAULT_POLL_TIMEOUT 500 


extern void die (char *error_message) ; 

extern int bind_port (int s, int port) ; 
extern int create_socket (int stype) ;
extern int forward_connect (int s, int port, char *address) ; 
extern int await_connection (int s) ; 
extern int start_listening (int s) ; 
extern int read_from_socket (int s, u_char *data) ;
extern int poll_socket (int s) ; 
extern int send_to_socket (int s, u_char *data, int buflen) ; 
extern int set_socket_buffers (int s) ;
extern int get_proto_number (const char *name) ;
extern int forward_socket_data (int s1, int s2) ;
extern int tie_sockets (int s1, int s2) ;
extern int __poll__ok_to_reply (int s) ; 


#define SOCK_STDIN 0 


#define MAX_SOCKETS 15  /* 16 sockets, wont need more than 2 thou */
#define SOCKETS_DESC_LEN 15 


#define SOCKSTATE_CLOSED	0
#define SOCKSTATE_LISTEN	1
#define SOCKSTATE_CONNECTED	2
#define SOCKSTATE_WAIT		4
#define SOCKSTATE_ACTIVE	8	/* active, but not in use (such as socket()ed) */

typedef int send_func (int, u_char *, int) ;
typedef int recv_func (int, u_char *) ;




/* polling */
#define	POLLMODE_OFF		0
#define POLLMODE_SERVER		1
#define	POLLMODE_CLIENT		2
#define POLLMODE_DEFAULT	POLLMODE_OFF
#define DEFAULT_POLL_INTERVAL	600000 	/* microseconds */

/* polling - protocol level */
#define POLLPROT_OFF		(int8_t)0x0
#define	POLLPROT_SPOLL		(int8_t)0x1 	/* server poll */
#define POLLPROT_CREPLY		(int8_t)0x2	/* client reply */

/* these values are to be ANDed with either SPOLL or CREPLYs */
#define	POLLPROT_DATA	        (int8_t)0x4	/* data follows .. */
#define POLLPROT_MORE_DATA	(int8_t)0x8	/* more data on stack  (FIXME: NOT IMPLEMENTED) */
#define	POLLPROT_NO_DATA	(int8_t)~POLLPROT_DATA	/* no data */

/* example:
 * server sends: POLLPROT_SPOLL | POLLPROT_DATA
 *               meaning that he's polling the client And sending data
 * client replies with: POLLPROT_CREPLY | POLLPROT_DATA | POLLPROT_MOREDATA
 *               indicating that the reply contains data, and theres more data on the stack
 * 
 * server sends: POLLPROT_SPOLL 
 * 		 just a poll, no data. 
 * client replies with: POLLPROT_CREPLY 
 * 		 the client does not have any data either.
 */



struct socket_struct { /* struct to keep track of our open sockets */
   int type ;
   int socket ; 
   char *desc ;
   int inuse ; 	    /* True = 1, False = 0 */
   int state ;
   int twinsocket ; /* the tunnel's other socket .. */
   int istunnel ;   /* use tunnel_*() functions on this socket */

   struct sockaddr *r_addr ; 
   
   
   /* protocol specific thingies */
   send_func *tunnel_send ;
   recv_func *tunnel_recv ;
   send_func *send ;
   recv_func *recv ;

   int tunnel_recv_type ;
   int tunnel_send_type ;
   int16_t tunnel_id ; 	/* each underlying protocol-tunneling code is to set this value somewhere in the packet */
   int16_t tunnel_rid ;	 
   
   /* polling */
   int poll_mode ;
   int poll_interval ;
   int8_t poll_key ;
   int poll_got_spoll ; /* client mode only */
   
   /* ICMP specific stuff (shouldnt really be here.. ) */
   int icmp_seq_val ; /* increment */
   int icmp_id_val ;
}  ;

   
extern struct socket_struct sockets[MAX_SOCKETS] ;

