/*
 *  tunnel.c - joint tunneling functions
 * 
 *  Some parts derived from Lars Brinkhoff's httptunnel
 * 
 * 
 * 
 * 
 * 
 * 
 * 
 * 
 * 
 */

#include "common.h"



int await_connection (int s)
{
   struct sockaddr addr;
   struct sockaddr_in *faddr ;   	

   socklen_t len;
   int t;

   
   
     len = sizeof(addr);
   
     t = accept (s, &addr, &len);
     if (t<0)
         die("await_connection: accept(2) failed") ;

	faddr = (struct sockaddr_in *) &addr ; 
   
   	fprintf(stderr, "await_connection:\tconnection from %s accepted. New socket fd is %i\n", inet_ntoa(faddr->sin_addr), t) ;
   
   sockets[s].state = SOCKSTATE_CONNECTED ;
   sockets[s].r_addr = &addr ; 

   
     return t;
}


int create_socket (int stype) 	/* simply creates a socket and returns the fd */
{
   struct protoent *proto ;
   int protonum ;
   int s=0 ;  	/* le soccet */
   char *ldesc ; 
   
   ldesc = (char *) malloc(SOCKETS_DESC_LEN + 1) ;


   switch(stype) {
    case SOCK_STREAM: /* tcp my friend */
      proto = getprotobyname((char *) "tcp") ;
      if (!proto)
	die("create_socket: getprotobyname(3) failed") ;
      
      protonum = proto->p_proto ;
      
      s = socket(AF_INET, SOCK_STREAM, protonum) ;
      if (s<0) 
	die("create_socket: socket(AF_INET, SOCK_STREAM, tcp) failed") ;
      
      fprintf(stderr, "create_socket:\t\tcreated AF_INET/SOCK_STREAM (tcp) socket attached to fd %i\n", s) ;


      sockets[s].type = IPPROTO_TCP ;
      
      sprintf(ldesc, "TCP") ; 

      
      
      break ;
    case SOCK_RAW:   /* icmp maybe.. */
      protonum = get_proto_number("icmp") ;
      
      s = socket(AF_INET, SOCK_RAW, protonum) ;
      if (s<0)
	die("create_Socket: socket(AF_INET, SOCK_RAW, icmp) failed") ;
      
      fprintf(stderr, "create_socket:\t\tcreated AF_INET/SOCK_RAW (icmp) socket attached fo fd %i\n", s) ; 

      sockets[s].type = IPPROTO_ICMP ;

      sprintf(ldesc, "ICMP") ; 

      

      break ;
      
    case SOCK_DGRAM:  /* udp..shoo */
      
      protonum = get_proto_number("udp") ;
      
      s = socket(AF_INET, SOCK_DGRAM, protonum) ;
      if (s<0)
	die("create_socket: socket(AF_INET, SOCK_DGRAM, udp) failed") ;
      
      fprintf(stderr, "create_socket:\t\tcreated AF_INET/SOCK_DGRAM (udp) socket attached to fd %i\n", s) ; 
      
      sockets[s].type = IPPROTO_UDP ; 
      
      sprintf(ldesc, "UDP") ; 
      
      break ;
      
    case SOCK_STDIN:	/* stdin .. */
      s = 0 ;
      sockets[s].type = 0 ;
      
      fprintf(stderr, "create_socket:\t\tbound socket 0 to stdin. sockets[0] structure set up\n") ; 

      sprintf(ldesc, "STDIN") ; 
      
      break ;
      
    default:
      fprintf(stderr, "create_socket:\t\tWARNING: unknown socket type: %i\n", stype) ; 
      return -1 ;
      

      
      break ; 
      
   }
   

   sockets[s].socket = s ; 
   sockets[s].state = SOCKSTATE_ACTIVE ;
   sockets[s].desc = ldesc ; 
   
   
   return s ; /* return socket fd */
}
   
   

int bind_port (int s, int port) 
{
   struct sockaddr_in addr ;

   socklen_t len ; 
   int t ; 

   
     addr.sin_family = AF_INET;
     addr.sin_port = htons ((short)port);
     addr.sin_addr.s_addr = INADDR_ANY;
   

   len = sizeof(addr) ; 
       
   t = bind(s, (struct sockaddr *)&addr, len) ;
   if (t<0) 
     die("bind_port: bind(2) failed") ; 
   
   
   fprintf(stderr, "bind_port:\t\tbound port %i to socket %i\n", (int)port, (int)s) ;

   
   return 1 ; 
   
}
   
    
int close_socket (int s)    
{
	sockets[s].inuse = 0 ;
   	sockets[s].type = 0 ; 
   	sockets[s].socket = -1 ;
   	sockets[s].desc = '\0' ;
   	sockets[s].state = SOCKSTATE_CLOSED ;
   	sockets[s].twinsocket = -1 ; 
   

   if (close(s))
     return 0 ;
   
   fprintf(stderr, "close_socket:\t\tWARNING: unable to close socket") ;
   perror("") ;

   return -1 ; 
}
 


int forward_connect (int s, int port, char *address) 
{
   
   struct sockaddr_in addr ;
   int ns = 0 ;
   
   addr.sin_family = AF_INET ;
   addr.sin_port = htons ((short)port) ;
   inet_aton(address, &addr.sin_addr) ;
   
   
   ns = connect(s, &addr, sizeof(addr)) ;
   if (ns<0) 
     die("forward_connect: connect(2) failed") ;
   
   fprintf(stderr, "forward_connect:\tns=%i\n", ns) ; 
   
   sockets[s].state = SOCKSTATE_CONNECTED ;
   

   sockets[s].r_addr = (struct sockaddr *) &addr ;
   
	return ns ;    	
   
}


int start_listening (int s) 
{
   int backlog = 2 ;
   int t = 0 ;
   
   
   
   t = listen(s, backlog) ;
   
   if (t<0) 
     die("start_listening: listen(2) failed") ; 
   
   fprintf(stderr, "start_listening:\tListening on socket %i\n", s) ; 

   
   sockets[s].state = SOCKSTATE_LISTEN ; 
   return 1; 
   
}


int read_from_socket (int s, u_char *data)
{

   	int r = 0 ;
   	int buflen = 0 ; 
   
   	buflen = READ_BUF_LEN; 


   	switch(sockets[s].type) {
	 case IPPROTO_TCP:
	   r = recv(s, data, buflen, 0) ;
	   break ;
	   
	 case IPPROTO_ICMP:
	   r = sockets[s].recv(s, data) ;
	   break ; 
	   
	 default:
	   r = read(s, data, buflen) ;
	   break ;
	}


   
       if (r<0) { /* error.. */
	  fprintf(stderr, "read_from_socket:\trecv(2)||read(2) failed: ") ;
	  perror(" ") ; 
	  return -1 ; 
       }
   
      if (r == 0) /* no data ..erp? */
     	return 0 ; 
   
   
   
	fprintf(stderr, "read_from_socket:\tread()==%i bytes\n", r) ;

   
   
   return r ; 
   
}

int send_to_socket (int s, u_char *data, int buflen)
{
   int r = 0 ;
   
   switch(sockets[s].type) {
    case IPPROTO_TCP:
      r = send(s, data, buflen, 0) ;
      break ;
      
    case IPPROTO_ICMP:
      r = sockets[s].send(s, data, buflen) ;
      
      break ; 
            
    default:
      r = write(s, data, buflen) ;
      break ;
   }
      
     if (r<0) {
	fprintf(stderr, "send_to_socket:\t\tSend(?) failed ") ;
	perror("") ;
	return -1 ; 
     }
   
   fprintf(stderr, "send_to_socket:\t\tSent %i bytes from %i byte buffer on socket %i\n", r, buflen, s) ;
   return r ;
}


int poll_socket (int s)
{

   struct pollfd  *fds ;
   int numfds = 0 ;
   int ret = 0  ;
   int timeout = DEFAULT_POLL_TIMEOUT ;
   
   
   
   
   
   fds = (struct pollfd *) calloc(1, sizeof(struct pollfd)) ;

   if (!fds)
     die("poll_socket: calloc(3) failed: Out of memory?") ; 
   
   
   fds->fd = s ; /* our socket */
   fds->events = POLLIN ;
   
   numfds = 1 ;
   
   
   
   ret = poll(fds, numfds, timeout) ;


   
   if (ret<0) { /* something went wrong */
      fprintf(stderr, "poll_socket:\t\tWARNING: poll(2) failed") ;
      perror("") ; 
      return -1 ; 
   }

   if (fds->revents == POLLHUP) { /* hangup .. */
      fprintf(stderr, "poll_socket:\t\tWARNING: peer hung up\n") ;
      return -1 ;
   }
   
   if (fds->revents == POLLERR) { /* error .. */
      fprintf(stderr, "poll_socket:\t\tWARNING: Unknown Error\n") ;
      return -1 ;
   }
   
   if (fds->revents == POLLNVAL) { /* socket not open */ 
      fprintf(stderr, "poll_socket:\t\tWARNING: socket fd %i not open\n", s) ;
      return -1 ; 
   }
   
   
   
   if (ret == 0) /* no data to be read */
     return 0 ; 
     
   
   /* theres data */
	fprintf(stderr, "poll_socket:\t\tfds->revents=%i on socket %i\n", fds->revents, s) ;
   
   
   return 1 ; 
   
     
   
}


int forward_socket_data (int s1, int s2) /* forward data betwen sockets (in simplex) (s2 is the encaps ..suck as ICMP socket) */
{
   int p = 0 ;
   int r =  0 ;
   int rs = 0 ; 
   int sone = 0 ;
   int stwo = 0 ;
   int i = 0  ;
   int last_poll = 0 ;
   int pollmode_socket = 0 ;
   int pollmode = 0 ;
   int ok_to_forward = 1 ;
   struct timeval tv ;

   u_char *databuf ; 
   u_char *clean_databuf  ;
   
   databuf = (u_char *) calloc(READ_BUF_LEN + 1, 1) ; 
   clean_databuf = databuf ; /* store a pointer to the beginning of databuf for later clearance */

   if (sockets[s1].poll_mode) { /* we can just poll on one socket.. (which is enuff... ) */
      pollmode_socket = s1 ;
      pollmode = sockets[pollmode_socket].poll_mode ;
   }
   else if (sockets[s2].poll_mode) {
      pollmode_socket = s2 ;
      pollmode = sockets[pollmode_socket].poll_mode ;
   }
   
   
   for ( i=0 ; i<2 ; i++ ) {

      
      if (i == 0) {
	 sone = s1 ; /* first run, socket one is s1 .. socket two is s2  */
	 stwo = s2 ;
      } else {
	 sone = s2 ; /* second run, socket one is s2, ... socket two is s1 */
	 stwo = s1 ; 
      }
      
      
      p = 0 ;
      
      databuf = clean_databuf ;
      memset(databuf, '\0', READ_BUF_LEN + 1) ;


      if (pollmode_socket>0) {  						/* if polling is enabled ..... */
	 sockets[pollmode_socket].poll_key = 0;
	 
	 if (pollmode == POLLMODE_SERVER) { 					/* we'll issue polls */
	    ok_to_forward = 1 ; 						/* so, we can send data at any time.. */

	    sockets[pollmode_socket].poll_key = POLLPROT_SPOLL  ; 		/* hence our default poll command is Server Poll.. */

	    if (gettimeofday(&tv, NULL) == 0) {
		if (tv.tv_usec > last_poll + sockets[pollmode_socket].poll_interval) {
		   send_to_socket(pollmode_socket, NULL, 0) ;
		   last_poll = tv.tv_usec ;
		}
	    } else
		die("forward_socket_data:\tgettimeofday() failed") ;
		

										/* if polling is enabled ... */
	 } else if (pollmode == POLLMODE_CLIENT) { 				/* we wont send 'til we're polled */
	    ok_to_forward = 0 ; 						/* we must wait for server poll before we can send our data .. */
	    
	    sockets[pollmode_socket].poll_key = POLLPROT_CREPLY ; 	        /* default poll command is Client Reply */
	 }
	 
	    

      }


	    
  	 p = poll_socket(sone) ;
	    
	    if ((__poll__ok_to_reply(pollmode_socket) == 0) && 
		(sockets[sone].twinsocket == pollmode_socket))
	       fprintf(stderr, "forward_socket_data:\t[POLLMODE]"
		      		"There is data on socket %i waiting to be sent."
		      		"Need a POLLPROT_SPOLL from peer first thou.\n",
		      		sone) ;
	 	else {
	        

	   if (p>0) { /* theres data to be received on socket sone */
	      if (sockets[sone].istunnel)
		 r = sockets[sone].tunnel_recv(sone, databuf) ;
	      else  
		r = read_from_socket(sone, databuf) ;


      
	      if (r>0) { /* there was data, so lets forward it to stwo */
		 	 /* however, if we're in poll mode, and we're the client.. 
			  * we'll hafta wait for a Server Poll first (which we might have already gotten) */

		 fprintf(stderr, "forward_socket_data:\t"
			 	 "Read %i bytes from socket %i\n",
			          r,
			 	  sone) ;
		   
		 if (sockets[stwo].istunnel)
		   rs = sockets[stwo].tunnel_send(stwo, databuf, r) ;
		 else
		   rs = send_to_socket(stwo, databuf, r) ;
		 
		 fprintf(stderr, "forward_socket_data:\t"
			 	 "Wrote %i bytes (from %i byte buffer) to socket %i\n",
			 	 rs, r, stwo) ;

	      } else if (r<0)  /* error, abort */
		 die("forward_to_socket:\tread_from_socket() or sockets.tunnel_recv() failed") ;
	      
		 
		 

		 if (rs<0) { /* error occured.. just inform */
		    die("forward_socket_data:\tsend_to_socket() or sockets.tunnel_send() failed") ;
/*		    fprintf(stderr, "forward_socket_data:\tsend_to_socket() failed. (wont abort)") ;
		    perror(" ") ; */
		 }
      	   } else if (p<0) { 	/* poll(2) error. Fatal for us */
	      die("forward_to_socket:\tpoll_socket() call failed.") ;
	   }
   
	   
      
		
      
		}
   }
   
   
   return 1 ; 
}
 



int tie_sockets (int s1, int s2)
{
      sockets[s1].twinsocket = s2 ;
      sockets[s2].twinsocket = s1 ;

   fprintf(stderr, "tie_sockets:\t\tSocket %i (%s) tied to socket %i (%s)\n", s1, sockets[s1].desc, s2, sockets[s2].desc) ;
   return 1 ; 
   
}

int __poll__ok_to_reply (int s)
{
   if (sockets[s].poll_mode == POLLMODE_CLIENT)
     if (sockets[s].poll_got_spoll > 0)
       return 1 ;
     else 
       return 0 ;
   else
     return 1 ;
   
}











int set_socket_buffers (int s) 
{
   int i, r, tcp ;
   
   tcp = get_proto_number("tcp") ;
   
   i = 1 ;
   r = 0 ;
	   
#ifdef SO_RCVLOWAT
   r = setsockopt(s, tcp, SO_RCVLOWAT, (void *)&i, sizeof(i)) ;

   if (r<0) {
      fprintf(stderr, "set_socket_buffers:\tsetsockopt(2) SO_RCVLOWAT failed") ;
      perror("") ;
      return -1 ;
   }

#endif
     
   
   return 1 ;
   
}
   

int get_proto_number (const char *name)
{
     struct protoent *p;
     int number;
   
     p = getprotobyname (name);
     if (!p)
         number = -1;
     else
         number = p->p_proto;
     endprotoent ();
   
     return number;
}



